/*
 * nlsop server code
 *
 * author: frank zirkelbach (frank.zirkelbach@physik.uni-augsburg.de)
 *
 * this program tries helping to understand the amorphous depuration
 * and recrystallization of SiCx while ion implantation at temperatures
 * below 400 degree celsius.
 * hopefully the program will simulate the stabilization of the
 * selforganizing lamella structure in the observed behaviour.
 *
 * refs: 
 *  - J. K. N. Lindner. Habil.Schrift, Universitaet Augsburg.
 *  - Maik Haeberlen. Diplomarbeit, Universitaet Augsburg.
 *
 * Copyright (C) 2004 Frank Zirkelbach
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

#include "nlsop.h"
#include "dfbapi.h"
#include "random.h"

#include "network.h"
#include "event.h"
#include "list.h"

#include "nlsop_general.h"

typedef struct s_priv {
  t_event event;
  t_net net;
  t_list client;
  t_list gui;
  t_list job;
} t_priv;

/* global */
int alert;
int gi;
t_priv priv;

/*
 * server specific stuff
 */

int usage(char *prog)
{
 puts("usage:");
 printf("%s <listen port>\n",prog);
 return 1;
}

int add_node(void) {

  int channel;
  unsigned char data;
  t_client client;
  int gui_chan;

  channel=network_manage_incoming(&(priv.net));
  if(channel==N_E_ACCEPT) {
    printf("accept failed!\n");
    return -1;
  }
  if(channel==N_E_MAXC) {
    printf("maximum connections reached!\n");
    return -1;
  }
  printf("connection from %s port %d (ch: %d)\n",
         priv.net.connection[channel].ip,priv.net.connection[channel].port,
         channel);

  /* are you client or gui? */
  network_receive_chan(&(priv.net),channel,&data,1);
  if(data==NLSOP_GUI) {
    gui_chan=channel;
    list_add_element(&(priv.gui),&gui_chan,sizeof(int));
    printf("node is a gui\n");
  }
  else if(data==NLSOP_CLIENT) {
    client.status=IDLE;
    client.channel=channel;
    list_add_element(&(priv.client),&client,sizeof(t_client));
    printf("node is a client\n");
  }
  else {
    printf("not a client or gui - lets kick that ass out of here!\n");
    network_close(&(priv.net),channel);
    return -1;
  }

  /* if we have a new node - care for it! */
  event_math(priv.net.connection[channel].fd,&(priv.event),READ,ADD);

  printf("\n");

  return 1;
}

int save_job(unsigned char dc) {

  char filename[128];
  int fd;
  int ret;
  d3_lattice d3l;
  info info;
  t_job *j;

  j=(t_job *)priv.job.current->data;

  printf("receiving data from client (#cells=%d)\n",j->size);

  network_receive_chan(&(priv.net),gi,(unsigned char *)&d3l,
                       sizeof(d3_lattice));
  network_receive_chan(&(priv.net),gi,(unsigned char *)&info,sizeof(info));
  network_receive_chan(&(priv.net),gi,j->ac,j->size*sizeof(unsigned char));
  network_receive_chan(&(priv.net),gi,(unsigned char *)j->cc,
                       j->size*sizeof(int));
  network_receive_chan(&(priv.net),gi,(unsigned char *)&(j->step),sizeof(int));

  if(dc!=DC_QUIT) {
    snprintf(filename,128,"./data/nlsop_b%f_c%f_s%f_ds%d_dr%f_-_%d_of_%d.save",
             j->info.b,j->info.c,j->info.s,
             j->info.diff_rate,j->info.dr_ac,
             j->step,j->info.steps);
    if((fd=open(filename,O_WRONLY|O_CREAT))<0) {
      printf("FATAL: unable to open file %s\n",filename);
      return -1;
    }

    if(write(fd,&d3l,sizeof(d3_lattice))<sizeof(d3_lattice)) {
      printf("FATAL: write of d3_lattice failed\n");
      return -1;
    }

    if(write(fd,&info,sizeof(info))<sizeof(info)) {
      printf("FATAL: write of info failed\n");
      return -1;
    }

    ret=write(fd,j->ac,j->size*sizeof(unsigned char));
    if(ret<j->size*sizeof(unsigned char)) {
      printf("FATAL: write of a/c states failed\n");
      return -1;
    }
   
    ret=write(fd,j->cc,j->size*sizeof(int));
    if(ret<j->size*sizeof(int)) {
      printf("FATAL: write of c.-conc. failed\n");
      return -1;
    }

    close(fd);
    
  }

  return 1;
}

int add_job(void) {

  t_job job;

  job.channel=-1;
  job.status=IN_QUEUE;
  job.progress=0;

  network_receive_chan(&(priv.net),gi,(unsigned char *)&(job.x),sizeof(int));
  network_receive_chan(&(priv.net),gi,(unsigned char *)&(job.y),sizeof(int));
  network_receive_chan(&(priv.net),gi,(unsigned char *)&(job.z),sizeof(int));
  network_receive_chan(&(priv.net),gi,(unsigned char *)&(job.info),
                       sizeof(info));

  job.size=job.x*job.y*job.z;

  job.ac=(unsigned char *)malloc(job.size*sizeof(unsigned char));
  if(job.ac==NULL) {
    printf("unable to malloc a/c memory\n");
    return -1;
  }

  job.cc=(int *)malloc(job.size*sizeof(int));
  if(job.cc==NULL) {
    printf("unable to malloc cc memory\n");
    return -1;
  }

  job.step=0;

  list_add_element(&(priv.job),&job,sizeof(t_job));
  
  printf("job added: b=%f | c=%f | s=%f ...\n",
         job.info.b,job.info.c,job.info.s);

  return 1;
}

int send_status(void) {

  unsigned char data;
  int count;
  int i;

  data=GUI_INFO;

  network_send_chan(&(priv.net),gi,&data,sizeof(unsigned char));
  printf("sending job + client list ...\n");

  count=list_count(&(priv.job));
  network_send_chan(&(priv.net),gi,(unsigned char *)&count,sizeof(int));
  list_reset(&(priv.job));
  for(i=0;i<count;i++) {
    network_send_chan(&(priv.net),gi,priv.job.current->data,sizeof(t_job));
    list_next(&(priv.job));
  }

  count=list_count(&(priv.client));
  network_send_chan(&(priv.net),gi,(unsigned char *)&count,sizeof(int));
  list_reset(&(priv.client));
  for(i=0;i<count;i++) {
    network_send_chan(&(priv.net),gi,priv.client.current->data,
                      sizeof(t_client));
    list_next(&(priv.client));
  }

  return 1;
}

int handle_node(void) {

  unsigned char data;
  t_client *c;
  t_job *j;

  for(gi=0;gi<MAX_CONNECTIONS;gi++) {
    if(FD_ISSET(priv.net.connection[gi].fd,&(priv.event.rfds))) {

      alert=0;
      alarm(1);
      network_receive_chan(&(priv.net),gi,&data,1);
      alarm(0);

      if(alert==1) return -1;

      if(list_search_data(&(priv.client),&gi,sizeof(int))==L_SUCCESS) {
        /* it's a client */
	list_search_data(&(priv.job),&gi,sizeof(int));
        j=(t_job *)priv.job.current->data;
        c=(t_client *)priv.client.current->data;

        if(data==DC_END) {
          save_job(DC_END);
          /* reset client */
          c->status=IDLE;
          /* free job memory */
          free(j->ac);
          free(j->cc);
          /* delete job entry */
          list_del_current(&(priv.job));
          printf("job ended, saved and removed from list.\n");
        }

        if(data==DC_OK) {
          save_job(DC_OK);
          /* inc progress state */
          j->progress+=1;
          printf("job at next level, saved.\n");
        }

        if(data==DC_QUIT) {
          save_job(DC_QUIT);
          /* network disconnect */
          event_math(priv.net.connection[gi].fd,&(priv.event),READ,REMOVE);
          network_close(&(priv.net),gi);
          /* del from client list */
          list_del_current(&(priv.client));
          /* change job state */
          j->status=IN_QUEUE;
          printf("client terminating, job queued, client removed.\n");
        }
      }

      else if(list_search_data(&(priv.gui),&gi,sizeof(int))==L_SUCCESS) {
        /* its a gui */
        if(data==GUI_ADDJOB) add_job();

        else if(data==GUI_INFO) send_status();

        else if(data==GUI_QUIT) {
          printf("disconnecting gui on channel %d\n",gi);
          event_math(priv.net.connection[gi].fd,&(priv.event),READ,REMOVE);
          network_close(&(priv.net),gi);
          list_del_current(&(priv.gui));
        }

        else {
          printf("unknown gui command\n");
          return -1;
        }
      }

      else {
        printf("this chan is not in client or gui list! i disconnect now!\n");
        event_math(priv.net.connection[gi].fd,&(priv.event),READ,REMOVE);
        network_close(&(priv.net),gi);
      }
    }
  }

  printf("\n");
   
  return 1;
}

int distribute_jobs(t_event *event,void *allineed) {

  int count_j,count_c,min;
  t_job *j;
  t_client *c;
  unsigned char data;
  d3_lattice d3l;

  list_reset(&(priv.job));
  list_reset(&(priv.client));

  count_j=0;
  count_c=0;

  if((priv.client.current==NULL)||(priv.job.current==NULL)) return 2;

  j=(t_job *)priv.job.current->data;
  c=(t_client *)priv.client.current->data;
  if(j->status==IN_QUEUE) count_j++;
  if(c->status==IDLE) count_c++;

  while(list_next(&(priv.job))!=L_NO_NEXT_ELEMENT) {
    j=(t_job *)priv.job.current->data;
    if(j->status==IN_QUEUE) count_j++;
  }
  while(list_next(&(priv.client))!=L_NO_NEXT_ELEMENT) {
    c=(t_client *)priv.client.current->data;
    if(c->status==IDLE) count_c++;
  }
 
  min=(count_c<count_j)?count_c:count_j;

  if(min!=0) {
    printf("d: distributing jobs ...\n");
    printf("%d queued jobs, %d idle clients\n\n",count_j,count_c);
  }

  list_reset(&(priv.job));
  list_reset(&(priv.client));

  while(min) {
    j=(t_job *)priv.job.current->data;
    c=(t_client *)priv.client.current->data;
    while(c->status!=IDLE) {
      list_next(&(priv.client));
      c=(t_client *)priv.client.current->data;
    }
    while(j->status!=IN_QUEUE) {
      list_next(&(priv.job));
      j=(t_job *)priv.job.current->data;
    }

    /* direct current job to current client */
    if(j->step==0) data=NLSOP_NJOB;
    else data=NLSOP_CJOB;

    c->status=WORK;
    j->channel=c->channel;
    j->status=IN_WORK;

    d3l.max_x=j->x;
    d3l.max_y=j->y;
    d3l.max_z=j->z;

    network_send_chan(&(priv.net),c->channel,&data,sizeof(unsigned char));
    network_send_chan(&(priv.net),c->channel,(unsigned char *)&d3l,
                      sizeof(d3_lattice));
    network_send_chan(&(priv.net),c->channel,(unsigned char *)&(j->info),
                      sizeof(info));

    if(data==NLSOP_CJOB) {
      network_send_chan(&(priv.net),c->channel,j->ac,
                        j->size*sizeof(unsigned char));
      network_receive_chan(&(priv.net),c->channel,&data,sizeof(unsigned char));
      network_send_chan(&(priv.net),c->channel,(unsigned char *)&(j->cc),
                        j->size*sizeof(int));
      network_receive_chan(&(priv.net),c->channel,&data,sizeof(unsigned char));
      network_send_chan(&(priv.net),c->channel,(unsigned char *)&(j->step),
                        sizeof(int));
      network_receive_chan(&(priv.net),c->channel,&data,sizeof(unsigned char));
    }

    --min;
    list_next(&(priv.client));
    list_next(&(priv.job));
  }

  return 1;
}

int parse_incoming(t_event *event,void *allineed) {

  /* decide what to do */
  if(FD_ISSET(priv.net.l_fd,&(priv.event.rfds))) {
    /* new node */
    printf("new node ...\n");
    add_node();
  }
  else {
    /* client/gui interaction */
    printf("node interaction ...\n");
    handle_node();
  }
    
  return 1;
}

void destroy_it(int signum) {

  printf("connection to client (ch %d) fucked up!\n",gi);
  event_math(priv.net.connection[gi].fd,&(priv.event),READ,REMOVE);
  network_close(&(priv.net),gi);
  if(list_search_data(&(priv.client),&gi,sizeof(int))==L_SUCCESS) {
    list_del_current(&(priv.client));
    printf("removed client from list\n");
  }
  if(list_search_data(&(priv.gui),&gi,sizeof(int))==L_SUCCESS) {
    list_del_current(&(priv.gui));
    printf("removed gui from list\n");
  }

  alert=1;
  alarm(0);

}

/*
 * main program
 */

int main(int argc,char **argv)
{

  int port;

  /* default values */
  port=1025;

  /* parse argv */
  if(argc==2) port=atoi(argv[1]);

  /* event init */
  event_init(&(priv.event),1);
  /* 10 sec event timeout - distributing jobs */
  event_set_timeout(&(priv.event),10,0);

  /* list init */
  list_init(&(priv.client),1);
  list_init(&(priv.gui),1);
  list_init(&(priv.job),1);

  /* connect to server */
  network_init(&(priv.net),1);
  network_set_listen_port(&(priv.net),port);
  if(network_listen(&(priv.net))!=N_SUCCESS) {
    printf("unable to listen on port %d, aborting!\n",port);
    return -1;
  }

  /* install sighandler */
  signal(SIGALRM,destroy_it);

  /* wait for events :) */
  event_math(priv.net.l_fd,&(priv.event),READ,ADD);
  printf("\nNLSOP_SERVER started!\n\n");
  event_start(&(priv.event),NULL,parse_incoming,distribute_jobs);

  return 1;
}

